@@ -549,48 +549,6 @@ install?(:local, :comm, :man) do
   end
 end
 
-# :stopdoc:
-module Gem
-  if defined?(Specification)
-    remove_const(:Specification)
-  end
-
-  class Specification < OpenStruct
-    def initialize(*)
-      super
-      yield(self) if defined?(yield)
-      self.executables ||= []
-      self.date ||= RUBY_RELEASE_DATE
-    end
-
-    def self.load(path)
-      src = File.open(path, "rb") {|f| f.read}
-      src.sub!(/\A#.*/, '')
-      eval(src, nil, path)
-    end
-
-    def to_ruby
-        <<-GEMSPEC
-Gem::Specification.new do |s|
-  s.name = #{name.dump}
-  s.version = #{version.dump}
-  s.date = #{date.dump}
-  s.summary = #{summary.dump}
-  s.description = #{description.dump}
-  s.homepage = #{homepage.dump}
-  s.authors = #{authors.inspect}
-  s.email = #{email.inspect}
-  s.files = #{files.inspect}
-end
-        GEMSPEC
-    end
-
-    def self.unresolved_deps
-      []
-    end
-  end
-end
-
 module RbInstall
   module Specs
     class FileCollector
@@ -599,7 +557,7 @@ module RbInstall
       end
 
       def collect
-        ruby_libraries + built_libraries
+        (ruby_libraries + built_libraries).sort
       end
 
       private
@@ -750,6 +708,21 @@ install?(:ext, :comm, :gem) do
   end
 end
 
+install?(:ext, :comm, :gem) do
+  require 'pathname'
+  gem_dir = Gem.default_dir
+  directories = Gem.ensure_gem_subdirectories(gem_dir, :mode => $dir_mode)
+  prepare "bundle gems", gem_dir, directories
+  Dir.glob(srcdir+'/gems/*.gem').each do |gem|
+    Gem.install gem, Gem::Requirement.default, :install_dir => with_destdir(Gem.dir), :domain => :local, :ignore_dependencies => true
+    gemname = Pathname(gem).basename
+    puts "#{" "*30}#{gemname}"
+  end
+  # fix directory permissions
+  # TODO: Gem.install should accept :dir_mode option or something
+  File.chmod($dir_mode, *Dir.glob(with_destdir(Gem.dir)+"/**/"))
+end
+
 parse_args()
 
 include FileUtils
@@ -757,18 +730,23 @@ include FileUtils::NoWrite if $dryrun
 @fileutils_output = STDOUT
 @fileutils_label = ''
 
+all = $install.delete(:all)
 $install << :local << :ext if $install.empty?
-$install.each do |inst|
+installs = $install.map do |inst|
   if !(procs = $install_procs[inst]) || procs.empty?
     next warn("unknown install target - #{inst}")
   end
-  procs.each do |block|
-    dir = Dir.pwd
-    begin
-      block.call
-    ensure
-      Dir.chdir(dir)
-    end
+  procs
+end
+installs.flatten!
+installs.uniq!
+installs |= $install_procs[:all] if all
+installs.each do |block|
+  dir = Dir.pwd
+  begin
+    block.call
+  ensure
+    Dir.chdir(dir)
   end
 end
 
